/* 
 * contrast_brightness.c 
 * version 1.1
 * author Felix Burkhardt
 * last change:  May 10 23:46:34 MEST 1998
 *
 * Routinen fue Kontrast- und Brightness-Veraenderungen.
 *
 * Welche Kanaele veraendert werden sollen, wird durch den
 * Parameter flag angegeben, wobei die Bits jewils fuer
 * R, G, B oder Alpha stehen. (also flag=7 waere RGB, flag=2 nur G)
 *
 * Die Routinen nehem einen Pointer auf Beschreibungstukturen
 * fuer die Anderungen
 * Jede Struktur besteht aus Reihenangabe und Veraenderungsfaktor
 * Die Strukturen MUESSEN nach aufsteigenden Reihen sortiert sein
 * Der Faktor ist um 0.5 skaliert (zwischen 0 and 1).
 *
 * Um den Kontrast zu veraendern wird der Mittelwert des Kanals
 * berechnet und als Kontrastgrenze verwendet.
 */

#include "contrast_brightness.h"
#include <stdio.h>

/* check overflow                                                          */
#define CHECK_MAX(x) (x>MAXVAL ? MAXVAL : x)
#define CHECK_MIN(x) (x<0 ? 0 : x)
#define CHECK_BORDER_MAX(x, b) (x>b ? b  : x)
#define CHECK_BORDER_MIN(x, b) (x<b ? b : x)

int changeContrastRows (long *pix_array,	/* inputChannel */
			int minrow, int maxrow,		/* min und Maxrow  */
			int width,
			int len,
			double start_val,	/* start-val contrast-change */
			double end_val,		/* end-val contrast-change */
			int flag,	/* which channels to change */
			int border_r, int border_g, /* decisionborders */
			int border_b, int border_a
);
int changeBrightnessRows (long *pix_array,	/* inputChannel */
			  int minrow, int maxrow,	/* min und Maxrow */
			  int width,
			  double start_val,	/* start-val brightness-change */
			  double end_val,	/* end-val brightness-change */
			  int flag	/* which channels to change */
);

extern pixval *changeContrast (pixval * pix_array_in,	/* inputPixmap */
		int width, int height,
		int changeDescrNum,	/* number of change-descriptors */
		changeDescriptorP cDPointer,	/* change-descriptors */
		int flag)
{				/* flag for channels  */
      pixval 
	*pix_array_out;
      int 
	len, i, ok,
	minrow, maxrow,
	border_r, border_g, /* decisionborders */
	border_b, border_a;
      double 
	start_val, end_val;

      len = width * height;


      pix_array_out = (pixval *) malloc (CHANNELS * len * sizeof (pixval));
      memcpy (pix_array_out, pix_array_in, len * CHANNELS);

      if (flag & CHANNEL_RED)
	    border_r = getMeanArray ((long *) pix_array_in, len, CHANNEL_RED);
      else border_r = MAXVAL;
      if (flag & CHANNEL_GREEN)
	    border_g = getMeanArray ((long *) pix_array_in, len, CHANNEL_GREEN);
      else border_g = MAXVAL;
      if (flag & CHANNEL_BLUE)
	    border_b = getMeanArray ((long *) pix_array_in, len, CHANNEL_BLUE);
      else border_b = MAXVAL;
      if (flag & CHANNEL_ALPHA)
	    border_a = getMeanArray ((long *) pix_array_in, len, CHANNEL_ALPHA);
      else border_a = MAXVAL;

      for (i = 0; i < changeDescrNum; i++) {
	    /* if there's only one field */
	    if (changeDescrNum == 1) {
		  minrow = 0;
		  maxrow = height;
		  start_val = cDPointer[0].val;
		  end_val = start_val;
	    }
	    /* if it's the last field and ends in the picture change the last rows */
	    else if (i == changeDescrNum - 1 && cDPointer[i].row < height) {
		  minrow = cDPointer[i].row;
		  maxrow = height;
		  start_val = cDPointer[i].val;
		  end_val = start_val;
	    }
	    /* if it's the last field there's nothing to be done */
	    else if (i == changeDescrNum - 1 && cDPointer[i].row >= height) {
		  break;
	    }
	    /* if we're in the field */
	    else {
		  /* if it's the first field and starts in the picture change the first rows */
		  if (i == 0 && cDPointer[0].row != 0) {
			minrow = 0;
			maxrow = cDPointer[0].row;
			start_val = cDPointer[i].val;
			end_val = start_val;
			ok = changeContrastRows ((long *) pix_array_out, minrow, maxrow,
				      width, len, start_val, end_val, flag, border_r, 
						 border_g, border_b, border_a);
		  }
		  minrow = cDPointer[i].row;
		  if ((maxrow = cDPointer[i + 1].row) > height)
			maxrow = height;
		  start_val = cDPointer[i].val;
		  end_val = cDPointer[i + 1].val;
	    }
	    ok = changeContrastRows ((long *) pix_array_out, minrow, maxrow,
				     width, len, start_val, end_val, flag, border_r, 
						 border_g, border_b, border_a);
      }

      return pix_array_out;
}

extern pixval *changeBrightness (pixval * pix_array_in,	/* inputPPM */
		  int width, int height,
		  int changeDescrNum,	/* number of change-descriptors   */
		  changeDescriptorP cDPointer,	/* change-descriptors */
		  int flag){			/* flag for channels  */
      pixval 
	*pix_array_out;
      int 
	len, 
	i, ok,
	minrow, maxrow;
      double 
	start_val, end_val;

      len = width * height;

      pix_array_out = (pixval *) malloc (CHANNELS * len * sizeof (pixval));
      memcpy (pix_array_out, pix_array_in, len * CHANNELS);


      for (i = 0; i < changeDescrNum; i++) {
	    /* if there's only one field */
	    if (changeDescrNum == 1) {
		  minrow = 0;
		  maxrow = height;
		  start_val = cDPointer[0].val;
		  end_val = start_val;
	    }
	    /* if it's the last field and ends in the picture change the last rows */
	    else if (i == changeDescrNum - 1 && cDPointer[i].row < height) {
		  minrow = cDPointer[i].row;
		  maxrow = height;
		  start_val = cDPointer[i].val;
		  end_val = start_val;
	    }
	    /* if it's the last field there's nothing to be done */
	    else if (i == changeDescrNum - 1 && cDPointer[i].row >= height) {
		  break;
	    }
	    /* if we're in the field */
	    else {
		  /* if it's the first field and starts in the picture change the first rows */
		  if (i == 0 && cDPointer[0].row != 0) {
			minrow = 0;
			maxrow = cDPointer[0].row;
			start_val = cDPointer[i].val;
			end_val = start_val;
			ok = changeBrightnessRows ((long *) pix_array_out, minrow, maxrow,
					   width, start_val, end_val, flag);
		  }
		  minrow = cDPointer[i].row;
		  if ((maxrow = cDPointer[i + 1].row) > height)
			maxrow = height;
		  start_val = cDPointer[i].val;
		  end_val = cDPointer[i + 1].val;
	    }
	    ok = changeBrightnessRows ((long *) pix_array_out, minrow, maxrow,
				       width, start_val, end_val, flag);
      }
      return pix_array_out;
}

/* actually change the contrast */
pixval changeC (pixval val, int border, int change)
{
      if (change > 0)		/* enhance contrast */
	    if (val >= border)
		  return CHECK_MAX (val + change);
	    else
		  return CHECK_MIN (val - change);
      else
	    /* reduce contrast */ if (val >= border)
	    return CHECK_BORDER_MIN (val + change, border);
      else
	    return CHECK_BORDER_MAX (val - change, border);
      return 0;
}

int changeContrastRows (long *pix_array,	/* pixChannel */
		    int minrow, int maxrow,	/* min und Maxrow */
		    int width,
		    int len,
		    double start_val,	/* start-val contrast-change */
		    double end_val,	/* end-val brightness-change */
		    int flag,		 /* which channels to change */
			int border_r, int border_g, /* decisionborders */
			int border_b, int border_a) {
      int 
	contr_change_r, contr_change_g, contr_change_b, contr_change_a,
	row, col,
	rowNum, actRow;
      long 
	rgba;
      pixval 
	r, g, b, a;
      double 
	start_fac, end_fac, step_fac, act_fac;

      rowNum = maxrow - minrow;
      start_fac = (start_val - 0.5) * 2;	/* scale factor around 0 */
      end_fac = (end_val - 0.5) * 2;	/* scale factor around 0         */

      step_fac = (double) (end_fac - start_fac) / rowNum;
      act_fac = start_fac;


      for (row = minrow; row < maxrow; row++) {

	    /* actualize change */
		  contr_change_r = (MAXVAL - border_r) * act_fac;
		  contr_change_g = (MAXVAL - border_g) * act_fac;
		  contr_change_b = (MAXVAL - border_b) * act_fac;
		  contr_change_a = (MAXVAL - border_a) * act_fac;
		  
		  actRow = row * width;

	    for (col = 0; col < width; col++) {

		  rgba = pix_array[(actRow + col)];

		  r = LONG_GETR (rgba);
		  g = LONG_GETG (rgba);
		  b = LONG_GETB (rgba);
		  a = LONG_GETA (rgba);

		  r = changeC (r, border_r, contr_change_r);
		  g = changeC (g, border_g, contr_change_g);
		  b = changeC (b, border_b, contr_change_b);
		  a = changeC (a, border_a, contr_change_a);

		  rgba = 0;

		  LONG_PUTRGBA (rgba, r, g, b, a);

		  pix_array[(actRow + col)] = rgba;
	    }
	    act_fac += step_fac;
      }
      return 0;
}

pixval changeB (pixval val, int change) {
      if (change > 0)		/* enhance brightness */
	    return CHECK_MAX (val + change);
      else			/* reduce brightness               */
	    return CHECK_MIN (val + change);
}

int changeBrightnessRows (
		      long *pix_array,	/* pixChannel */
		      int minrow, int maxrow,	/* min und Maxrow */
		      int width,
		      double start_val,	/* factor for contrast-change */
		      double end_val,	/* factor for contrast-change */
		      int flag ) {
      int 
	row, col, bright_change, rowNum, actRow;
      double 
	start_fac, end_fac, step_fac, act_fac;
      long 
	rgba;
      pixval 
	r, g, b, a;

      rowNum = maxrow - minrow;
      start_fac = (start_val - 0.5) * 2;	/* scale factor around 0 */
      end_fac = (end_val - 0.5) * 2;	/* scale factor around 0 */

      step_fac = (double) (end_fac - start_fac) / rowNum;
      act_fac = start_fac;


      for (row = minrow; row < maxrow; row++) {

	    bright_change = MAXVAL * act_fac;
	    actRow = row * width;

	    for (col = 0; col < width; col++) {
		  rgba = pix_array[(actRow + col)];
		  r = LONG_GETR (rgba);
		  g = LONG_GETG (rgba);
		  b = LONG_GETB (rgba);
		  a = LONG_GETA (rgba);

		  if (flag & CHANNEL_RED)
			r = changeB (r, bright_change);
		  if (flag & CHANNEL_GREEN)
			g = changeB (g, bright_change);
		  if (flag & CHANNEL_RED)
			b = changeB (b, bright_change);
		  if (flag & CHANNEL_RED)
			a = changeB (a, bright_change);

		  rgba = 0;

		  LONG_PUTRGBA (rgba, r, g, b, a);

		  pix_array[(actRow + col)] = rgba;
	    }
	    act_fac += step_fac;
      }
      return 0;
}
